#include <assert.h>
#include <gmp.h>

#define NDEBUG
#undef DEBUG
#undef GC_DEBUG

#if defined(GC_DEBUG) || defined(DEBUG)
#include <stdio.h>
#endif

#define CAML_NAME_SPACE
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/custom.h>
#include <caml/camlidlruntime.h>


extern void camlidl_mpz_ptr_ml2c(value, mpz_ptr *);
extern value camlidl_mpz_ptr_c2ml(mpz_ptr *);
extern void camlidl_mpq_ptr_ml2c(value, mpq_ptr *);
extern value camlidl_mpq_ptr_c2ml(mpq_ptr *);

#include <yices_c.h>

#define CAMLID(id) ml_yices_##id
#define CAMLCALL(id) value CAMLID(id)

#ifdef __GNUC__
#define field(x) .x =
#else
#define field(x)
#endif

static inline yices_context context_of_value(value r) {
  return * ((yices_context *) (Data_custom_val(r)));
}

static inline value value_of_expr(yices_expr e) {
  return (value) e;
}

static inline yices_expr expr_of_value(value e) {
  return (yices_expr) e;
}

static inline value value_of_type(yices_type e) {
  return (value) e;
}

static inline yices_type type_of_value(value e) {
  return (yices_type) e;
}

static inline value value_of_var_decl(yices_var_decl e) {
  return (value) e;
}

static inline yices_type var_decl_of_value(value e) {
  return (yices_var_decl) e;
}

static inline yices_model model_of_value(value e) {
  return (yices_model) e;
}

static inline value value_of_model(yices_model e) {
  return (value) e;
}

static inline value value_of_lbool(lbool v) {
  switch(v) {
  case l_true: return 1;
  case l_false: return 3;
  default: return 5;
  }
}

void CAMLID(context_finalize)(value r) {
  /* TODO */
  yices_del_context (context_of_value(r));
#ifdef GC_DEBUG
  fputs("Yices.context finalized\n", stderr);
#endif
}

struct custom_operations CAMLID(custom_context) = {
    field(identifier) "Yices.context",
    field(finalize) CAMLID(context_finalize),
    field(compare) NULL,
    field(hash) NULL,
    field(serialize) NULL,
    field(deserialize) NULL,
};

CAMLCALL(mk_context)(value dummy) {
  CAMLparam0();
  CAMLlocal1(r);
  r = caml_alloc_custom(& CAMLID(custom_context),
			       sizeof(yices_context), 0, 1);
  * ((yices_context *) (Data_custom_val(r))) = yices_mk_context();
  CAMLreturn(r);
}

CAMLCALL(dump_context)(value context) {
  CAMLparam1(context);
  yices_dump_context(context_of_value(context));
  CAMLreturn(Val_unit);
}

CAMLCALL(mk_true)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_expr(yices_mk_true(context_of_value(context))));
}

CAMLCALL(mk_false)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_expr(yices_mk_false(context_of_value(context))));
}

CAMLCALL(assert_expr)(value context, value expr) {
  CAMLparam2(context, expr);
  yices_assert(context_of_value(context), expr_of_value(expr));
  CAMLreturn(Val_unit);
}

CAMLCALL(check)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_lbool(yices_check(context_of_value(context))));
}

CAMLCALL(push)(value context) {
  CAMLparam1(context);
  yices_push(context_of_value(context));
  CAMLreturn(Val_unit);
}

CAMLCALL(pop)(value context) {
  CAMLparam1(context);
  yices_pop(context_of_value(context));
  CAMLreturn(Val_unit);
}

CAMLCALL(mk_bool_var)(value context, value name) {
  CAMLparam2(context, name);
  CAMLreturn(value_of_expr(yices_mk_bool_var(context_of_value(context), String_val(name))));
}

CAMLCALL(mk_fresh_bool_var)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_expr(yices_mk_fresh_bool_var(context_of_value(context))));
}

unsigned CAMLID(list_length)(value list) {
  unsigned length = 0;
  while(list != 1) {
   list = Field(list, 1);
    length++;
  }
  return length;
}

void CAMLID(expr_list_to_array)(value list, yices_expr* t, unsigned n) {
  for(unsigned i=0; i<n; i++) {
    assert(list != 1);
    t[i] = expr_of_value(Field(list, 0));
    list = Field(list, 1);
  }
  assert(list == 1);
}

#define mk_list_operator(op) \
CAMLCALL(mk_##op)(value context, value list) { \
  CAMLparam2(context, list); \
  int n = CAMLID(list_length)(list); \
  yices_expr args[n]; \
  CAMLID(expr_list_to_array)(list, args, n); \
  CAMLreturn(value_of_expr(yices_mk_##op(context_of_value(context), args, n))); \
}

mk_list_operator(and)
mk_list_operator(or)
mk_list_operator(sum)
mk_list_operator(sub)
mk_list_operator(mul)

CAMLCALL(mk_not)(value context, value expr) {
  CAMLparam2(context, expr);
  CAMLreturn(value_of_expr(yices_mk_not(context_of_value(context), expr_of_value(expr))));
}

CAMLCALL(mk_ite)(value context, value cond, value expr1, value expr2) {
  CAMLparam4(context, cond, expr1, expr2);
  CAMLreturn(value_of_expr(yices_mk_ite(context_of_value(context), expr_of_value(cond), expr_of_value(expr1), expr_of_value(expr2))));
}

#define mk_comparison(op) \
CAMLCALL(mk_##op)(value context, value a1, value a2) { \
  CAMLparam3(context, a1, a2); \
  CAMLreturn(value_of_expr(yices_mk_##op(context_of_value(context), \
    expr_of_value(a1), expr_of_value(a2)))); \
}

mk_comparison(lt)
mk_comparison(gt)
mk_comparison(le)
mk_comparison(ge)
mk_comparison(eq)
mk_comparison(diseq)

CAMLCALL(mk_type)(value context, value type_name) {
  CAMLparam2(context, type_name);
  CAMLreturn(value_of_type(yices_mk_type(context_of_value(context), String_val(type_name))));
}

CAMLCALL(mk_var_decl)(value context, value name, value type) {
  CAMLparam3(context, name, type);
  CAMLreturn(value_of_var_decl(yices_mk_var_decl(context_of_value(context), String_val(name), type_of_value(type))));
}

CAMLCALL(mk_bool_var_decl)(value context, value name) {
  CAMLparam2(context, name);
  CAMLreturn(value_of_var_decl(yices_mk_bool_var_decl(context_of_value(context), String_val(name))));
}

CAMLCALL(mk_var_from_decl)(value context, value var_decl) {
  CAMLparam2(context, var_decl);
  CAMLreturn(value_of_expr(yices_mk_var_from_decl(context_of_value(context), var_decl_of_value(var_decl))));
}

CAMLCALL(mk_num)(value context, value n) {
  CAMLparam2(context, n);
  CAMLreturn(value_of_expr(yices_mk_num(context_of_value(context), Int_val(n))));
}

CAMLCALL(mk_num_from_string)(value context, value n) {
  CAMLparam2(context, n);
  CAMLreturn(value_of_expr(yices_mk_num_from_string(context_of_value(context), String_val(n))));
}

CAMLCALL(mk_num_from_mpz)(value context, value _v_n) {
  CAMLparam2(context, _v_n);
  mpz_ptr n;
  camlidl_mpz_ptr_ml2c(_v_n, &n);

#ifdef DEBUG
  gmp_printf("mk_num_from_mpz(_, %Zd)\n", n);
  fflush(stdout);
#endif
  CAMLreturn(value_of_expr(yices_mk_num_from_mpz(context_of_value(context), n)));
}

CAMLCALL(get_model)(value context) {
  CAMLparam1(context);
  CAMLreturn(value_of_model(yices_get_model(context_of_value(context))));
}

CAMLCALL(get_value)(value model, value decl) {
  CAMLparam2(model, decl);
  CAMLreturn(value_of_lbool(yices_get_value(model_of_value(model), var_decl_of_value(decl))));
}

CAMLCALL(get_mpq_value)(value model, value decl) {
  CAMLparam2(model, decl);
  mpq_t r;
  mpq_ptr r_p = r;
  mpq_init(r);
  yices_get_mpq_value(model_of_value(model), var_decl_of_value(decl), r);
#ifdef DEBUG
  gmp_printf("get_mpq_value = %Qd\n", r);
#endif
  CAMLreturn(camlidl_mpq_ptr_c2ml(& r_p));
}
